/*
* Original code from SOTA game project.
*
* Reference:
*  https://shroudoftheavatar.com/forum/index.php?threads/game-object-replace-tool.3954/
*
* Addition of Distribution modes
*  Offset (relative)
*  Absolute placement (global coords)
*  Random (between two vectors)
*  Absolute Random (Absolute global coords between two vectors)
*
* Multiple object replacers
*
* Allowance for object scattering with and without replacement object(s)
*  Good use for quick scatter across terrain of proxy objects which can
*  later be batch replaced with the replacement object(s)
*
*/
using UnityEngine;
using UnityEditor;
using System.Linq;
using System.Collections.Generic;
 
 
 
 
// Per transform option to apply
public enum Mode
{
	NoChange,
	Offset,
	Absolute,
	Randomized,
	RandomAbsolute
}
 
 
 
 
// How to distribute the object list
// For an "even" distribution == 1/n objects
public enum Distribution
{
	Weighted,
	WeightedCountdown,
	Randomized,
	RandomWeighted
}
 
 
 
 
// Support container for each type of transform (Pos, Rot, Scle)
public class Blob
{
	private Mode mode;
	private Vector3 min;
	private Vector3 max;
	private string m_label = "";
 
 
 
 
	public Blob(string label, Vector3 defaultValue = new Vector3())
	{
		m_label = label;
		min = defaultValue;
	}
 
 
 
 
	// Convience abstraction
	public void OnGUI()
	{
		// Set values
		var desc = "No Change\n\n";
		desc += "Offset will transform the object(s) relative to the entered vector\n\n";
		desc += "Absolute will contrain the object(s) to the exact entered vector\n\n";
		desc += "Randomized will take randomly pick between the highs and lows and offset the object\n\n";
		desc += "Random Absolute keeps objects within the min/max absolute vector";
 
 
 
 
		GUIContent transPop = new GUIContent(m_label, desc);
		mode = (Mode)EditorGUILayout.EnumPopup(transPop, mode);
		if(mode != Mode.NoChange)
		{
			var rand = (mode == Mode.Randomized || mode == Mode.RandomAbsolute);
			var sublabel = (rand) ? "Min" : "";
			min = EditorGUILayout.Vector3Field(sublabel, min);
			if(rand)
			{
				max = EditorGUILayout.Vector3Field("Max", max);
			}
		}
	}
 
 
 
 
	// Consolidate the vector(s) into a new result based on the mode
	public Vector3 Adjust(Vector3 orig)
	{
		Vector3 result = min;
		switch (mode)
		{
			case Mode.Absolute:
				break;
 
 
 
 
			case Mode.RandomAbsolute:
				result.x = Random.Range(min.x, max.x);
				result.y = Random.Range(min.y, max.y);
				result.z = Random.Range(min.z, max.z);
				break;
 
 
 
 
			case Mode.Randomized:
				result.x = orig.x + Random.Range(min.x, max.x);
				result.y = orig.y + Random.Range(min.y, max.y);
				result.z = orig.z + Random.Range(min.z, max.z);
				break;
 
 
 
 
			case Mode.Offset:
				result = orig + min;
				break;
 
 
 
 
			default:
				result = orig;
				break;
		}
		return result;
	}
}
 
 
 
 
// A tool that will replace all instances of one game object (based on name or selection) in a scene with another
// Note: Save your scene before using this tool!!!!!!
public class GameObjectReplaceTool : EditorWindow
{
	// The target proxy object to replace
	private GameObject m_targetGameObject = null;
 
 
 
 
	// The object(s) used to replace the source proxy
	private GameObject m_replacerGameObject = null;
	private IList<GameObject> m_replacerGameObjects = new List<GameObject>();
	private IList<float> m_replacerWeights = new List<float>();
	private Distribution distrib = Distribution.Weighted;
	private Blob m_position = new Blob("Position");
	private Blob m_rotation = new Blob("Rotation");
	private Blob m_scale = new Blob("Scale", new Vector3(1, 1, 1));
	//private int randomSeed = 8675309;
	private bool useObjectSelection = false;
	private bool maintainChildren = false;
	private bool maintainParent = true;
 
 
 
 
	[MenuItem("SOTA/Game Object Replacer")]
	static void Init()
	{
		// Create window
		GameObjectReplaceTool window = (GameObjectReplaceTool)EditorWindow.GetWindow(typeof(GameObjectReplaceTool));
		window.minSize = new Vector2(400, 500);
	}
 
 
 
 
	//
	public void OnGUI()
	{
		if(!useObjectSelection)
		{
			// The target and source
			GUIContent targetPop = new GUIContent("Target Object", "Select or drag an object onto this field that will be replaced by all of the Target Objects");
			m_targetGameObject = EditorGUILayout.ObjectField(targetPop, m_targetGameObject, typeof(GameObject), true) as GameObject;
		}
 
 
 
 
		if(m_replacerGameObject)
		{
			m_replacerGameObjects.Add(m_replacerGameObject);
			m_replacerWeights.Add(1.0f);
			m_replacerGameObject = null;
		}
 
 
 
 
		GUIContent replacerPop = new GUIContent("Replacer Object", "Select or drag an object onto this field that will designate all sources used to replaced the Target Object (based on name)");
		for(int i=0; i < m_replacerGameObjects.Count; i++)
		{
			EditorGUILayout.BeginHorizontal();
			if(distrib != Distribution.Randomized)
			{
				m_replacerWeights[i] = EditorGUILayout.FloatField(replacerPop, m_replacerWeights[i]);
				m_replacerGameObjects[i] = EditorGUILayout.ObjectField(m_replacerGameObjects[i], typeof(GameObject), true) as GameObject;
			}
			else
			{
				m_replacerGameObjects[i] = EditorGUILayout.ObjectField(replacerPop, m_replacerGameObjects[i], typeof(GameObject), true) as GameObject;
			}
			EditorGUILayout.EndHorizontal();
		}
		m_replacerGameObject = EditorGUILayout.ObjectField(replacerPop, m_replacerGameObject, typeof(GameObject), true) as GameObject;
 
 
 
 
 
 
 
 
		GUIContent selectPop = new GUIContent("Use Selected Objects", "Check to use only selected objects within the scene as the targets for replacement");
		useObjectSelection = EditorGUILayout.Toggle(selectPop, useObjectSelection);
 
 
 
 
		/*
		 * Random is not random enough -- irony.
		if(distrib != Distribution.Weighted && distrib != Distribution.WeightedCountdown)
		{
			GUIContent randomPop = new GUIContent("Random Seed", "Enter a value or preserve a value for consistant results");
			randomSeed = EditorGUILayout.IntField(randomPop, randomSeed);
		}
		*/
 
 
 
 
		var desc = "Weighted will use the weights in first in/first out order (Round Robin)\n\n";
		desc += "Weighted Countdown will use the weights with priority order (highest weight to lowest)\n\n";
		desc += "Randomized replaces target object or selection with random object from replacer object(s) list\n\n";
		desc += "Random Weighted replaces target object or selection with a random object from the replacer object(s) list based on the weights";
 
 
 
 
		GUIContent distroPop = new GUIContent("Distribution", desc);
		distrib = (Distribution)EditorGUILayout.EnumPopup(distroPop, distrib);
 
 
 
 
		// Build Transform Controls
		m_position.OnGUI();
		m_rotation.OnGUI();
		m_scale.OnGUI();
 
 
 
 
		GUIContent childPop = new GUIContent("Maintain Children", "Check to replace only the Source's parent object and place all of the source's children into the Target Object");
		maintainChildren = EditorGUILayout.Toggle(childPop, maintainChildren);
 
 
 
 
		GUIContent parentPop = new GUIContent("Keep Target's Parent", "Check to place Target Objects into the same spot as their Sources within the hierarchy");
		maintainParent = EditorGUILayout.Toggle(parentPop, maintainParent);
 
 
 
 
		// Hit the button to do the swap
		if(GUILayout.Button("Replace Objects"))
		{
			CheckForReplace();
		}
	}
 
 
 
 
	//
	private void CheckForReplace()
	{
		// Early bail if we are missing a target and at least 1 source
		if(!useObjectSelection && m_targetGameObject == null)
		{
			EditorGUILayout.HelpBox("You must have at least one object to replace!", MessageType.Error);
			Debug.LogWarning("GameObject Replace Tool: target must be specified!");
			return;
		}
 
 
 
 
		// clean up the reference list
		m_replacerGameObjects = (from t in m_replacerGameObjects
							  where t != null
							  select t).ToList();
 
 
 
 
		var targets = new List<GameObject>();
 
 
 
 
		// Cull the target object list depending on distribution setting
		// Use Selected will be the explict scene selections
		if(useObjectSelection)
		{
			targets = (from t in Selection.transforms
					  where t.gameObject != null
					  select t.gameObject).ToList();
 
 
 
 
			if(targets.Count == 0)
			{
				EditorGUILayout.HelpBox("You must have at least one object to replace!", MessageType.Error);
				Debug.LogWarning("GameObject Replace Tool: target must be specified!");
				return;
			}
		}
		else
		{
			string thisTargetName = m_targetGameObject.name;
 
 
 
 
			// Get all the objects we need to "replace"
			targets = (from t in GameObject.FindObjectsOfType<GameObject>()
					  where t.name == thisTargetName
					  select t).ToList();
		}
 
 
 
 
		// make sure we have something to transform, even if no replacement objects are setup
		IList<GameObject> objectReplacements = targets;
		IList<float> objectReplacementWeights = (from t in Enumerable.Range(0, targets.Count)
												 select 1.0f).ToList();
 
 
 
 
		// If there are replacement objects....
		if(m_replacerGameObjects.Count > 0)
		{
			objectReplacements = m_replacerGameObjects;
			objectReplacementWeights = m_replacerWeights;
		}
 
 
 
 
		// precalc for curve weight
		var weightCount = objectReplacementWeights.Count;
		var weightTotal = objectReplacementWeights.Sum();
		var weightScale = weightTotal * targets.Count / weightCount;
		var scale = targets.Count / weightTotal;
 
 
 
 
		// get a running list of # replacer objects to use per weight
		Dictionary<int, float> weights = new Dictionary<int, float>();
		for(int i=0; i < weightCount; i++)
		{
			// slight cheat for now, weights will go a little higher, sue me.
			weights.Add(i, objectReplacementWeights[i] * scale);
		}
 
 
 
 
		// check distribution options -- weight or randomized
		// N.B. a single "replacer" will only get used once no matter the option
		// Process the proxy targets
		//Random.seed = randomSeed;
 
 
 
 
		// the full weight range for all replacers
 
 
 
 
		for(int index=0; index < targets.Count; index++)
		{
			var replacerIndex = 0;
			if(distrib == Distribution.Randomized && weightCount > 0)
			{
				// simple random noise across all the possible replacers
				replacerIndex = Random.Range(0, weightCount);
			}
			else if(distrib == Distribution.RandomWeighted)
			{
				float rand = Random.Range(0.0f, weightScale);
				for(var i=0; i < weightCount; i++)
				{
					if(rand < weights[i])
					{
						replacerIndex = i;
						break;
					}
					rand -= weights[i];
				}
			}
			else if(distrib == Distribution.Weighted)
			{
				// used like a circular queue/round robin
				replacerIndex = index % weightCount;
				while(true)
				{
					if(weights[replacerIndex] > 0.0f)
					{
						weights[replacerIndex] -= 1;
						break;
					}
					replacerIndex = (replacerIndex + 1) % weightCount;
				}
			}
			else if(distrib == Distribution.WeightedCountdown)
			{
				replacerIndex = weights.Aggregate((l, r) => l.Value > r.Value ? l : r).Key;
				weights[replacerIndex] -= 1;
			}
 
 
 
 
			var replacerObject = objectReplacements[replacerIndex];
			ReplaceObject(targets[index], replacerObject);
		}
	}
 
 
 
 
	//
	private void ReplaceObject(GameObject thisObj, GameObject replacerObject)
	{
		// Register for undo
		Undo.RecordObject(thisObj, "Replace Target");
 
 
 
 
		// Get transform data
		Vector3 p = m_position.Adjust(thisObj.transform.position);
		Vector3 r = m_rotation.Adjust(thisObj.transform.eulerAngles);
		Vector3 s = m_scale.Adjust(thisObj.transform.localScale);
 
 
 
 
		// Create a clone based on prefab status
		GameObject source = null;
		PrefabType thisObjectType = PrefabUtility.GetPrefabType(replacerObject);
		if(thisObjectType != PrefabType.None)
		{
			Object thisPrefab = replacerObject;
			if(thisObjectType != PrefabType.Prefab)
			{
				// For some reason, IntantiatePrefab needs this...
				thisPrefab = PrefabUtility.GetPrefabParent(replacerObject);
			}
			source = PrefabUtility.InstantiatePrefab(thisPrefab) as GameObject;
		}
		else
		{
			source = Instantiate(replacerObject) as GameObject;
 
 
 
 
			// Clear out the extra (clone) appended to name
			source.name = replacerObject.name;
		}
 
 
 
 
		// Set transform
		source.transform.position = p;
		source.transform.eulerAngles = r;
		source.transform.localScale = s;
 
 
 
 
		Undo.RegisterCreatedObjectUndo(source, "Replace Source");
 
 
 
 
		// Check children stuffs
		if(maintainChildren)
		{
			foreach(Transform thisChild in thisObj.transform)
			{
				thisChild.parent = source.transform;
			}
		}
 
 
 
 
		// Check for parents
		if(maintainParent)
		{
			source.transform.parent = thisObj.transform.parent;
		}
 
 
 
 
		// Undo stuffs
		Undo.DestroyObjectImmediate(thisObj);
		Undo.RecordObject(source, "Replace Source");
	}
}